/* Sweep v1.07 (clone of the Rutgers "slide" program)
   Last Updated: 09/12/97
   Authors: Tom Dyas              tdyas@usacs.rutgers.edu
            Manish Vachharajani   mvachhar@usacs.rutgers.edu
   Public Domain with the "It's all your fault!" guarantee.

   Here at Rutgers, the "slide" program grants root access to
   authorized people by checking to see if they are in the "slide"
   user group. Thus, an administrator does not need to give out the
   root password to lots of people. In fact, the root account can be
   disabled providing even more security. (This presupposes that the
   passwords of anyone with root access are secure by themselves.)

   Sweep should run on any Unix-compatible system. However, it was
   designed to compile under Linux first. The makefile may need
   modification if you compile under any other type of Unix.

   Change Log
   ----------
   050395 TKD 0.91  - The need for using a change log esp. when multiple
      people will be updating this code has become blatantly obvious.

   050395 MV  0.92  - Updated program to include a new form of subshell which
      will not destroy any existing environment and will run the user's
      default shell.

   050495 TKD 1.00  - Rewrote the entire program to clean it up and enforce
      the clean coding practice. Comments galore have been added. I feel that
      sweep now has enough features to qualify as a version 1.00 program.

   051095 TKD 1.02  - Removed the code section that removes lone dots in the
      PATH. Not feasible since startup files change the PATH and add it
      sometimes anyways. The default is to not run the root account shell.
      This can now be configured with the ROOTMODE macro.

   060995 MV  1.03  - Fixed a small bug that wouldn't allow you to suspend a 
      swept shell. Now the '-' character is only prepended to the argv[0] 
      of a spawned shell when rootmode = 1.

   091295 TKD 1.04  - Fixed bug which would cause denial of a root shell
      if /etc/shells did not exist or was invalid. Various documentation
      cleanups.

   091395 TKD 1.05  - Lets log everything and anything!
   091395 TKD 1.06  - Random access denied messages.
   092695 MV  1.06a - Added compile time switch for the no suser sub-shells.
          TKD       - Changed random number generation routines. 
   100595 TKD 1.06b - Makefile now compiles sweep as a static binary.
   112695 TKD 1.06c - Did not copy username out of static data area.
   030897 MV  1.07  - Fixed a small bug that caused seg faults.
*/

#define USE_BSD

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>
#include <time.h>
#include <errno.h>

/*
 *  configuration macros - modify these to customize sweep for any
 *    particular installation.
 */

#define VERSION        "1.07"
#define GROUP          "sweep"
#define LAST_RESORT    "/bin/sh"
#define SHELLS_PATH    "/etc/shells"
#define PATH           "/sbin:/usr/sbin:/usr/local/sbin:/bin:/usr/bin:"\
                       "/usr/local/bin:/usr/ucb:/usr/X11/bin"
#define LOG_FACILITY   LOG_AUTH
#define LOG_LEVEL      LOG_INFO
#define ROOTMODE       0
#define ALREADY_SUSER  0

/*
 *  global variables
 */

int    rootmode;        /* TRUE if the user wants a full root login */
const char * progname;  /* argv[0] for this copy of sweep */
const char * user;      /* name of the current user */
char * env[2];          /* temporary environment space  */


/*
 * error reporting macros (makes life simpler down below)
 */

#define ERROR(str) {\
syslog(LOG_LEVEL, str ", user %s", user);\
fprintf(stderr, "%s: " str "\n", progname);\
closelog();\
exit(1);\
}

#define SYSCALL_ERROR(str) {\
int tmp1 = errno; char * tmp2 = strerror(tmp1);\
syslog(LOG_LEVEL, str ": (%d) %s", tmp1, tmp2);\
fprintf(stderr, "%s: " str ": (%d) %s\n",progname,tmp1,tmp2);\
closelog();\
exit(1);\
}


/*
 *  implementation of changing access denied messages
 */

const char * no_access_messages[] = 
{
  "Sorry, not now dear.",
  "access denied",
  "You want root? You can't handle root!",
  "You'll get root when the sky falls.",
  "Life is like a box of chocolates but you still don't get root.",
};
#define NUM_NO_ACCESS_MSGS (sizeof(no_access_messages)/sizeof(const char *))

const char * no_access_msg(void)
{
  return(no_access_messages[time(NULL) % NUM_NO_ACCESS_MSGS]);
}


/* Checks to see if the user is really a member of the given root
 * access group. This routine returns:
 *
 *  0 - Affirmative, user is a member of the group.
 *  1 - Negative, user is NOT a member of the group.
 *  2 - Negative, the group does not exist.
 */
int ingroup(const char * group_name, const char * user_name)
{
  struct group * gp;
  gid_t groups[30], numgroups;
  int i, j;

  /* Check #1: Information on the access group must exist. */
  if ((gp = getgrnam(group_name)) == NULL) return 2;

  /* Check #2: User must be listed in /etc/group ONCE. */
  i = j = 0;
  while (*(gp->gr_mem+j))
    {
      if (strcmp(*(gp->gr_mem+j), user_name) == 0) i++;
      j++;
    }
  if (i != 1) return 1;

  /* Check #3: User must have the gid of the access group. */
  numgroups = getgroups(30,groups);
  for (i = 0; i < numgroups; i++)
    {
      if (groups[i] == gp->gr_gid) return 0;
    }
  return 1;
}

/* Check to see if the user if authorized to use sweep. The return
 * values are the same as the ingroup() function.
 */
int authorize(const char * user_name)
{
  int result = ingroup(GROUP,user_name);
  return result;
}


/* Convert an array of strings into a single string. */
char * combine(int numstrs, char **strs)
{
  int i, size;
  char *buffer;

  /* Determine the size of the entire string. */
  size = numstrs;
  for (i = 0; i < numstrs; i++) size += strlen(strs[i]);

  /* Allocate space for the string. Exit on any error. */
  buffer = (char *) malloc( size * sizeof(char) );
  if (buffer == NULL)
    ERROR("memory allocation error");

  /* Build a single string from the array of strings. */
  strcpy(buffer,strs[0]);
  for (i = 1; i < numstrs; i++)
    {
      strcat(buffer," ");
      strcat(buffer,strs[i]);
    }

  /* Return the completed string. */
  return buffer;
}

/* Return nonzero if the given shell is valid. */
int validshell(const char * name)
{
  FILE * fp;
  char line[1024];

  /* Open the /etc/shells file for read access. Exit on any error. */
  if ((fp = fopen(SHELLS_PATH,"r")) == NULL) return 0;

  /* Scan the file line-by-line looking for a match. */
  while (fgets(line,sizeof(line),fp) != NULL)
    {
      if (line[strlen(line)-1] == '\n') line[strlen(line)-1] = '\0';
      if (strcmp(line,name) == 0)
	{
	  fclose(fp);
	  return 1;
	}
    }
  fclose(fp);
  return 0;
}

/* Determine the correct argv[0] for a shell name. */
char * shell_name(const char * shell)
{
  char * name;
  if (strrchr(shell,'/') != NULL)
    {
      if ((name = strdup(shell)) == NULL)
	ERROR("memory allocation error");
      name = strrchr(name,'/');
      
      if (rootmode)
	name[0] = '-';
      else
	name++;
    }
  else
    {
      name = (char *) malloc((2 + strlen(shell)) * sizeof(char));
      if (name == NULL)
	ERROR("memory allocation error");
      
      if (rootmode)
	{
	  strcpy(name,"-");
	  strcat(name,shell);
	}
      else
	strcpy(name,shell);
    }
  return name;
}

/* Reset the environment to that of a root account. */
void rootenv(struct passwd * pp)
{
  char *old_TERM, *old_DISPLAY;

  old_TERM = getenv("TERM");
  old_DISPLAY = getenv("DISPLAY");

  env[0] = "PATH=" PATH;
  env[1] = NULL;
/*  environ = env; */

  if (old_TERM != NULL)
  if (setenv("TERM",old_TERM,1) < 0)
    ERROR("not enough memory for environment");

  if (old_DISPLAY != NULL)
  if (setenv("DISPLAY",old_DISPLAY,1) < 0)
    ERROR("not enough memory for environment");

  if (setenv("USER",pp->pw_name,1) < 0)
    ERROR("not enough memory for environment");

  if (setenv("LOGNAME",pp->pw_name,1) < 0)
    ERROR("not enough memory for environment");

  if (setenv("HOME",pp->pw_dir,1) < 0)
    ERROR("not enough memory for environment");

  if (setenv("SHELL",pp->pw_shell,1) < 0)
    ERROR("not enough memory for environment");
}

/* Run a subshell using whatever shell and environment is called for. */
void rootshell(const char * command)
{
  struct passwd * pp;
  char *name, *shell;

  /* Get information on the superuser account. */
  shell = NULL;
  if ((pp = getpwuid((uid_t)0)) == NULL)
    {
      syslog(LOG_LEVEL,"no superuser account, using " LAST_RESORT);
      fprintf(stderr,"%s: no superuser account, using " LAST_RESORT "\n",progname);
      shell = LAST_RESORT;
    }

  /* Determine and validate the exact shell to be run. */
  if (!shell) shell = getenv("SHELL");
  if (rootmode || !shell) shell = pp->pw_shell;
  if (!validshell(shell))
    {
      syslog(LOG_LEVEL, "unauthorized shell %s, using" LAST_RESORT, shell);
      fprintf(stderr,"%s: unauthorized shell, using " LAST_RESORT "\n",progname);
      shell = LAST_RESORT;
    }

  /* Reset the environment correctly for full root shells. */
  if (rootmode) rootenv(pp);

  /* Set the permissions to that of the superuser. */
  if (setgid(pp->pw_gid) < 0)
    SYSCALL_ERROR("setgid");

  if (initgroups(pp->pw_name,pp->pw_gid) < 0)
    SYSCALL_ERROR("initgroups");

  if (setuid((uid_t)0) < 0)
    SYSCALL_ERROR("setuid");

  /* Run a root shell with the command if present. */
  name = shell_name(shell);

  if (strlen(command) == 0)
    {
      closelog();
      execl(shell,name,NULL);
      fprintf(stderr,"%s: execl: %s\n",progname,strerror(errno));
      exit(1);
    }

  /* Otherwise, run the shell with just the command. */
  closelog();
  execl(shell,name,"-c",command,NULL);
  fprintf(stderr,"%s: execl: %s\n",progname,strerror(errno));
  exit(1);
}

/* Display a quick usage screen. */
void usage(void)
{
  printf("usage: %s [-rnvh] [command]\n",progname);
  printf("       -r   run a full root shell\n");
  printf("       -n   run the current user's shell\n");
  printf("       -v   display version information\n");
  printf("       -h   this usage screen\n");
}

/* Display version information. */
void version(void)
{
  printf("%s v" VERSION " - secure root shell access\n\n",progname);
  printf("Compiled: " __TIME__ " on " __DATE__ "\n\n");
  printf("Authors:\n");
  printf("  Tom Dyas              tdyas@usacs.rutgers.edu\n");
  printf("  Manish Vachharajani   mvachhar@usacs.rutgers.edu\n");
}

/* Scan the command line parameters for options. */
void cmdline(int argc, char **argv, char **command)
{
  int i, j;

  /* Parse each command line option. More than one allowed per switch. */
  for (i = 1; i < argc; i++)
    {
      if (argv[i][0] != '-') break;
      for (j = 1; j < strlen(argv[i]); j++)
	{
	  switch (argv[i][j])
	    {
	    case 'v':
	      syslog(LOG_LEVEL,"version option activated, user %s",user);
	      version();
	      closelog();
	      exit(0);
	      break;
	    case 'h':
	      syslog(LOG_LEVEL,"usage option activated, user %s",user);
	      usage();
	      closelog();
	      exit(0);
	      break;
	    case 'r':
	      rootmode = 1;
	      break;
            case 'n':
              rootmode = 0;
              break;
	    default:
	      syslog(LOG_LEVEL,"unknown option '%c', user %s",argv[i][j],user);
	      usage();
	      closelog();
	      exit(0);
	      break;
	    }
	}
    }
  if (argc == i)
    *command = "";
  else
    *command = combine(argc - i, &argv[i]);
}

/* If you don't know what main() is, then learn C. :) */
int main(int argc, char *argv[])
{
  struct passwd * pp;
  char * command;

  /* Initalize global variables and logging system. */
  progname = argv[0];
  rootmode = ROOTMODE;
  openlog(progname, LOG_PID | LOG_NOWAIT, LOG_FACILITY);

  /* Prevent root from running another root shell. */
#if ALREADY_SUSER
  if (getuid() == 0)
    {
      syslog(LOG_LEVEL,"already superuser");
      fprintf(stderr,"%s: already superuser\n",progname);
      closelog();
      return 0;
    }
#endif /* ALREADY_SUSER */

  /* Get the information on the current user. */
  if ((pp = getpwuid(getuid())) == NULL)
    {
      syslog(LOG_LEVEL,"unknown user, uid = %d, euid = %d",getuid(),geteuid());
      fprintf(stderr,"%s: Who are you?\n",progname);
      closelog();
      return 1;
    }

  /* Copy the username out of the static data area. */
  if ((user = strdup(pp->pw_name)) == NULL)
      {
	syslog(LOG_LEVEL,"memory allocation error");
	fprintf(stderr,"%s: memory allocation error\n",progname);
	closelog();
	return 1;
      }

  /* Parse the command line arguments. */
  cmdline(argc,argv,&command);

  /* Check the user's access level and take action accordingly. */
  switch (authorize(user))
    {
    case 0:
      /* User can run a root shell. */
      if (rootmode)
	syslog(LOG_LEVEL,"success by %s (full root): %s",user,command);
      else
	syslog(LOG_LEVEL,"success by %s: %s",user,command);
      rootshell(command);
      break;

    case 1:
      /* User is NOT authorized to run a root shell. */
      syslog(LOG_LEVEL,"failure by %s: %s",user,command);
      fprintf(stderr,"%s: %s\n",progname,no_access_msg());
      break;

    case 2:
      /* Group does not exist. */
      syslog(LOG_LEVEL,"group " GROUP " does not exist");
      fprintf(stderr,"%s: group " GROUP " does not exist\n",progname);
      break;

    default:
      /* Should not happen. Unknown own return code. */
      syslog(LOG_LEVEL,"unknown return code from authorize()");
      fprintf(stderr,"%s: unknown return code from authorize()\n",progname);
      break;
    }
  closelog();
  return 1;
}
